package com.arms.api.issue.almapi.service;

import static com.arms.api.util.alm.지라유틸.*;

import com.arms.api.issue.almapi.model.dto.*;
import com.arms.api.issue.almapi.model.entity.AlmIssueEntity;
import com.arms.api.issue.almapi.model.vo.지라이슈_데이터;
import com.arms.api.issue.almapi.repository.지라이슈_저장소;
import com.arms.api.serverinfo.model.서버정보_데이터;
import com.arms.api.util.alm.지라API_정보;
import com.arms.api.util.alm.지라유틸;
import com.arms.egovframework.javaservice.esframework.erros.response.ErrorCode;
import com.arms.api.util.errors.ErrorLogUtil;
import com.arms.api.util.response.CommonResponse;

import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpStatus;
import org.apache.http.client.utils.DateUtils;
import org.springframework.stereotype.Component;
import org.springframework.web.reactive.function.client.WebClient;
import org.springframework.web.reactive.function.client.WebClientResponseException;

import java.util.*;

@Slf4j
@Component
@AllArgsConstructor
public class 클라우드_지라_이슈전략 implements 이슈전략 {

    private final 지라유틸 지라유틸;
    private final 지라API_정보 지라API_정보;
    private final 지라이슈_저장소 지라이슈_저장소;

    @Override
    public List<지라이슈_데이터> getIssueList(서버정보_데이터 서버정보, String projectKeyOrId) {

        int startAt = 0;
        final int maxResults = 지라API_정보.getParameter().getMaxResults();
        boolean isLast = false;

        WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

        List<지라이슈_데이터> issueList = new ArrayList<>();

        while (!isLast) {
            String endpoint = 지라API_정보.getEndpoint().getSearch()
                    + "?jql=project=" + projectKeyOrId
                    + "&startAt=" + startAt
                    + "&maxResults=" + maxResults
                    + "&" + 지라API_정보.getParameter().getFields();

            지라이슈조회_데이터 response = this.fetchIssueList(webClient, endpoint, 서버정보, projectKeyOrId);
            issueList.addAll(response.getIssues());

            isLast = this.checkIfLast(response.getTotal(), issueList.size());
            startAt += maxResults;
        }

        return issueList;
    }

    private 지라이슈조회_데이터 fetchIssueList(WebClient webClient, String endpoint, 서버정보_데이터 서버정보, String projectKeyOrId) {
        지라이슈조회_데이터 response = 지라유틸.get(webClient, endpoint, 지라이슈조회_데이터.class).block();

        if (response == null || response.getIssues() == null) {
            String errorMessage = String.format("클라우드 지라[%s] :: 프로젝트_키_또는_아이디[%s] :: getIssueList 조회 목록이 없습니다.",
                                                            서버정보.getUri(), projectKeyOrId);
            log.error(errorMessage);
            throw new IllegalArgumentException(ErrorCode.이슈_조회_오류.getErrorMsg() + errorMessage);
        }

        return response;
    }

    private boolean checkIfLast(int totalSize, int currentSize) {
        return totalSize == currentSize;
    }

    @Override
    public 지라이슈_데이터 createIssue(서버정보_데이터 서버정보, AlmIssueCreationDTO almIssueCreationDTO) {

        IssueCreationFieldsDTO issueCreationFieldsDTO = almIssueCreationDTO.getFields();
        if (issueCreationFieldsDTO == null) {
            String errorMessage = String.format("%s :: %s 이슈 생성 필드 데이터가 존재 하지 않습니다.",
                                                서버정보.getType(), 서버정보.getUri());
            throw new IllegalArgumentException(ErrorCode.요청본문_오류체크.getErrorMsg() + " :: " + errorMessage);
        }

        String projectKeyOrId = null;
        String issueTypeId = null;

        if (issueCreationFieldsDTO.getProject() != null && StringUtils.isNotEmpty(issueCreationFieldsDTO.getProject().getId())) {
            projectKeyOrId = issueCreationFieldsDTO.getProject().getId();
        }

        if (issueCreationFieldsDTO.getIssuetype() != null && StringUtils.isNotEmpty(issueCreationFieldsDTO.getIssuetype().getId())) {
            issueTypeId = issueCreationFieldsDTO.getIssuetype().getId();
        }

        if (StringUtils.isEmpty(projectKeyOrId) || StringUtils.isEmpty(issueTypeId)) {
            String errorMessage = String.format("%s :: %s 이슈 생성 필드 확인에 필요한 프로젝트 아이디, 이슈유형 아이디가 존재하지 않습니다.",
                                                서버정보.getType(), 서버정보.getUri());
            throw new IllegalArgumentException(errorMessage);
        }

        WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

        /* ***
         * 프로젝트 와 이슈 유형에 따라 이슈 생성 시 들어가는 fields의 내용을 확인하는 부분
         *** */
        Map<String, CloudJiraIssueCreationFieldMetadata.FieldMetadata> fieldMetadataMap
                = 지라유틸.checkFieldMetadata(webClient, projectKeyOrId, issueTypeId);
        CloudJiraIssueFieldDTO fieldDTO = this.validateAndAddFields(issueCreationFieldsDTO, fieldMetadataMap, 서버정보, projectKeyOrId, issueTypeId);

        CloudJiraIssueCreationDTO issueCreationDTO = new CloudJiraIssueCreationDTO();
        issueCreationDTO.setFields(fieldDTO);

        String endpoint = 지라API_정보.getEndpoint().getIssue().getBase();
        지라이슈_데이터 returnIssueDTO;
        try {
            returnIssueDTO = 지라유틸.post(webClient, endpoint, issueCreationDTO, 지라이슈_데이터.class).block();
            log.info("{}[{}], 프로젝트 : {}, 이슈유형 : {}, 생성 필드 : {}, 이슈 생성하기",
                    서버정보.getType(), 서버정보.getUri(), projectKeyOrId, issueTypeId, issueCreationDTO.toString());
        }
        catch (Exception e) {
            String errorMessage = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                                    String.format("%s[%s],  프로젝트[%s], 이슈유형[%s], 생성 필드 :: 이슈 생성하기 중 오류",
                                                서버정보.getType(), 서버정보.getUri(), projectKeyOrId, issueTypeId, issueCreationDTO.toString()));
            throw new IllegalArgumentException(errorMessage);
        }

        if (returnIssueDTO == null) {
            String errorMessage = String.format("%s[%s], 프로젝트[%s], 이슈유형[%s], 생성 필드 :: {}, 생성된 이슈 데이터가 NULL 입니다.",
                                                서버정보.getType(), 서버정보.getUri(), projectKeyOrId, issueTypeId, issueCreationDTO.toString());
            log.error(errorMessage);
            throw new IllegalArgumentException(errorMessage);
        }

        return returnIssueDTO;
    }

    @Override
    public Map<String, Object> 이슈_수정하기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디, AlmIssueCreationDTO almIssueCreationDTO) {

        try {
            WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

            String endpoint = 지라API_정보.getEndpoint().getIssue().getBase() + "/" + 이슈_키_또는_아이디;
            Map<String, Object> 결과 = new HashMap<>();

            IssueCreationFieldsDTO 필드_데이터 = almIssueCreationDTO.getFields();

            CloudJiraIssueCreationDTO 수정_데이터 = new CloudJiraIssueCreationDTO();
            CloudJiraIssueFieldDTO 클라우드_필드_데이터 = new CloudJiraIssueFieldDTO();

            if (필드_데이터.getSummary() != null) {
                클라우드_필드_데이터.setSummary(필드_데이터.getSummary());
            }

            if (필드_데이터.getDescription() != null) {
                클라우드_필드_데이터.setDescription(내용_변환(필드_데이터.getDescription()));
            }

            if (필드_데이터.getLabels() != null) {
                클라우드_필드_데이터.setLabels(필드_데이터.getLabels());
            }

            수정_데이터.setFields(클라우드_필드_데이터);

            CommonResponse.ApiResult<?> 응답_결과 = 지라유틸.executePut(webClient, endpoint, 수정_데이터);

            if (응답_결과.isSuccess()) {
                결과.put("success", 응답_결과.isSuccess());
                결과.put("message", "이슈 수정 성공");
            }
            else {
                결과.put("success", 응답_결과.isSuccess());
                결과.put("message", 응답_결과.getError().getMessage());
            }

            return 결과;
        }
        catch (Exception e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "클라우드 지라("+ 서버정보.getUri() +") :: 이슈 키("+ 이슈_키_또는_아이디+ ") :: 이슈_수정하기에 실패하였습니다.");
            throw new IllegalArgumentException(ErrorCode.이슈수정_오류.getErrorMsg() + " :: " + 에러로그);
        }

    }

    @Override
    public Map<String, Object> 이슈_상태_변경하기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디, String 상태_아이디) {

        String 이슈전환_아이디 = null;
        try {
            WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

            String endpoint = 지라API_정보.getEndpoint().getIssue().getBase() + "/" + 이슈_키_또는_아이디 + "/transitions";
            Map<String, Object> 결과 = new HashMap<>();

            // Transition 조회
            이슈전환_아이디 = 이슈전환_아이디_조회하기(서버정보, 이슈_키_또는_아이디, 상태_아이디);

            // 이슈 상태 변경
            if (이슈전환_아이디 != null) {

                지라이슈전환_데이터.Transition 전환 = 지라이슈전환_데이터.Transition.builder()
                        .id(이슈전환_아이디)
                        .build();
                지라이슈전환_데이터 수정_데이터 = 지라이슈전환_데이터.builder()
                        .transition(전환)
                        .build();

                CommonResponse.ApiResult<?> 응답_결과 = 지라유틸.executePost(webClient, endpoint, 수정_데이터);

                if (응답_결과.isSuccess()) {
                    결과.put("success", 응답_결과.isSuccess());
                    결과.put("message", "이슈 상태 변경 성공");
                }
                else {
                    결과.put("success", 응답_결과.isSuccess());
                    결과.put("message", 응답_결과.getError().getMessage());
                }
            }
            else {
                String 에러로그 = "클라우드 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + 이슈_키_또는_아이디 + ") :: 상태 아이디(" + 상태_아이디 + ") :: 해당 업무 흐름으로 변경이 불가능 합니다.";
                log.error(에러로그);

                결과.put("success", false);
                결과.put("message", "변경할 이슈 상태가 존재하지 않습니다.");
            }

            return 결과;

        }
        catch (Exception e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "클라우드 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + 이슈_키_또는_아이디 + ") :: 상태 아이디(" + 상태_아이디 + ") :: 전환 아이디(" + 이슈전환_아이디 + ") :: 이슈_상태_변경하기에 실패하였습니다.");
            throw new IllegalArgumentException(ErrorCode.이슈전환_오류.getErrorMsg() + " :: " + 에러로그);
        }
    }

    public String 이슈전환_아이디_조회하기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디, String 상태_아이디) {

        try {
            WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

            String endpoint = 지라API_정보.getEndpoint().getIssue().getBase() + "/" + 이슈_키_또는_아이디 + "/transitions";

            지라이슈전환_데이터 이슈전환_데이터 = 지라유틸.get(webClient, endpoint, 지라이슈전환_데이터.class).block();

            return Optional.ofNullable(이슈전환_데이터)
                    .map(지라이슈전환_데이터::getTransitions)
                    .orElse(Collections.emptyList()).stream()
                    .filter(데이터 -> {
                        if (데이터.getTo() != null && !StringUtils.isBlank(데이터.getTo().getId())) {
                            return 상태_아이디.equals(데이터.getTo().getId());
                        }
                        return false;
                    })
                    .findFirst()
                    .map(지라이슈전환_데이터.Transition::getId)
                    .orElse(null);

        } catch (Exception e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "클라우드 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + 이슈_키_또는_아이디 + ") :: 상태 아이디(" + 상태_아이디 + ") :: 이슈전환_아이디_조회하기에 실패하였습니다.");
            throw new IllegalArgumentException(ErrorCode.이슈전환_조회_오류.getErrorMsg() + " :: " + 에러로그);
        }
    }

    @Override
    public Map<String, Object> 이슈_삭제하기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디) {

        try {
            WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

            boolean 하위이슈_삭제유무 = 지라API_정보.getParameter().isDeleteSubtasks();
            String endpoint = 지라API_정보.getEndpoint().getIssue().getBase() + "/" + 이슈_키_또는_아이디 + "?deleteSubtasks=" + 하위이슈_삭제유무;
            Map<String, Object> 결과 = new HashMap<>();

            CommonResponse.ApiResult<?> 응답_결과 = 지라유틸.executeDelete(webClient, endpoint);

            if (응답_결과.isSuccess()) {
                결과.put("success", 응답_결과.isSuccess());
                결과.put("message", "이슈 삭제 성공");
            }
            else {
                결과.put("success", 응답_결과.isSuccess());
                결과.put("message", 응답_결과.getError().getMessage());
            }

            return 결과;
        }
        catch (Exception e) {
            String 에러로그 = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "클라우드 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + 이슈_키_또는_아이디 + ") :: 이슈_수정하기에 실패하였습니다.");
            throw new IllegalArgumentException(ErrorCode.이슈수정_오류.getErrorMsg() + " :: " + 에러로그);
        }
    }

    public CloudJiraIssueFieldDTO.내용 내용_변환(String 입력_데이터) {

        CloudJiraIssueFieldDTO.콘텐츠_아이템 콘텐츠_아이템 = CloudJiraIssueFieldDTO.콘텐츠_아이템.builder()
                .text(입력_데이터)
                .type("text")
                .build();

        List<CloudJiraIssueFieldDTO.콘텐츠_아이템> 콘텐츠_아이템_리스트 = new ArrayList<>();
        콘텐츠_아이템_리스트.add(콘텐츠_아이템);

        CloudJiraIssueFieldDTO.콘텐츠 콘텐츠 = CloudJiraIssueFieldDTO.콘텐츠.builder()
                .content(콘텐츠_아이템_리스트)
                .type("paragraph")
                .build();

        List<CloudJiraIssueFieldDTO.콘텐츠> 콘텐츠_리스트 = new ArrayList<>();
        콘텐츠_리스트.add(콘텐츠);

        CloudJiraIssueFieldDTO.내용 내용 = CloudJiraIssueFieldDTO.내용.builder()
                .content(콘텐츠_리스트)
                .type("doc")
                .version(1)
                .build();

        return 내용;
    }

    @Override
    public 지라이슈_데이터 이슈_상세정보_가져오기(서버정보_데이터 서버정보, String issueKeyOrId) throws Exception {

        String endpoint = 지라API_정보.getEndpoint().getIssue().getBase() + "/" + issueKeyOrId + "?" + 지라API_정보.getParameter().getFields();

        WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

        try {
            지라이슈_데이터 almIssueData = 지라유틸.getWithRawData(webClient, endpoint, 지라이슈_데이터.class).block();
            return almIssueData;
        }
        catch (WebClientResponseException e) {
            ErrorLogUtil.exceptionLogging(e, this.getClass().getName(),
                    "클라우드 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + issueKeyOrId + ") :: 이슈_상세정보_가져오기에 실패하였습니다.");

            if (e.getStatusCode() != null && e.getStatusCode().value() == HttpStatus.SC_NOT_FOUND) {
                return null;
            }

            throw e;
        }
        catch (Exception e) {
            ErrorLogUtil.exceptionLogging(e, this.getClass().getName(),
                    "클라우드 지라(" + 서버정보.getUri() + ") :: 이슈 키(" + issueKeyOrId + ") :: 이슈_상세정보_가져오기에 실패하였습니다.");

            throw e;
        }
    }

    // 연결된 이슈 가져오기.
    @Override
    public List<지라이슈_데이터> 연결이슈_가져오기(서버정보_데이터 서버정보, String issueKeyOrId) {

        WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

        Set<String> issueIdsSet = new HashSet<>();
        issueIdsSet.add(issueKeyOrId);

        List<지라이슈_데이터> relationIssueList = new ArrayList<>();
        연결이슈_가져오기_재귀(webClient, 서버정보, issueKeyOrId, issueIdsSet, relationIssueList);

        return relationIssueList;
    }

    public void 연결이슈_가져오기_재귀(WebClient webClient, 서버정보_데이터 서버정보, String issueKeyOrId, Set<String> issueIdsSet, List<지라이슈_데이터> relationIssueList) {

        int startAt = 0;
        int maxResults = 지라API_정보.getParameter().getMaxResults();

        try {
            while (true) {
                String linkedIssueEndpoint = 지라API_정보.이슈키_대체하기(지라API_정보.getEndpoint().getIssue().getFull().getLinkedIssue(), issueKeyOrId);
                String endpoint = linkedIssueEndpoint + "&startAt=" + startAt + "&maxResults=" + maxResults;

                지라이슈조회_데이터 searchResultRelationIssue = 지라유틸.getWithRawData(webClient, endpoint, 지라이슈조회_데이터.class).block();

                if (searchResultRelationIssue == null) {
                    log.info("클라우드 지라({}) :: 이슈 키({}) :: 연결이슈 조회결과가 Null입니다.", 서버정보.getUri(), issueKeyOrId);
                    return;
                }
                else if (searchResultRelationIssue.getIssues() == null || searchResultRelationIssue.getIssues().size() == 0) {
                    log.info("클라우드 지라({}) :: 이슈 키({}) :: 연결이슈 목록이 없습니다.", 서버정보.getUri(), issueKeyOrId);
                    return;
                }

                for (지라이슈_데이터 relationIssue : searchResultRelationIssue.getIssues()) {

                    if (issueIdsSet.contains(relationIssue.getKey())) {
                        continue;
                    }

                    // 요구사항 이슈인 경우는 미수집
                    relationIssue.set서버아이디(서버정보.getConnectId());
                    String documentId = relationIssue.조회조건_아이디();
                    AlmIssueEntity almIssueEntity = 지라이슈_저장소.findRecentDocById(documentId);

                    if (almIssueEntity != null && Boolean.TRUE.equals(almIssueEntity.getIsReq())) {
                        log.info("클라우드 지라 ({}) :: 연결이슈_아이디 :: {} :: 요구사항 이슈입니다."
                                , 서버정보.getUri(), almIssueEntity.getKey());
                        continue;
                    }

                    issueIdsSet.add(relationIssue.getKey());
                    relationIssueList.add(relationIssue);
                    연결이슈_가져오기_재귀(webClient, 서버정보, relationIssue.getKey(), issueIdsSet, relationIssueList);
                }

                startAt += maxResults;
                if (startAt >= searchResultRelationIssue.getTotal()) {
                    break;
                }
            }
        }
        catch (Exception e) {
            ErrorLogUtil.exceptionLogging(e, this.getClass().getName(),
                    "클라우드 지라("+ 서버정보.getUri()  +") :: 이슈 키("+ issueKeyOrId+ ") :: 링크드 이슈 조회에 실패하였습니다.");
        }
    }

    @Override
    public List<지라이슈_데이터> 서브테스크_가져오기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디) throws Exception {

        WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

        int 검색_시작_지점 = 0;
        int 최대_검색수 = 지라API_정보.getParameter().getMaxResults();

        List<지라이슈_데이터> 서브테스크_목록 = new ArrayList<>(); // 이슈 저장

        try {
            while (true) {
                String subTaskEndpoint = 지라API_정보.이슈키_대체하기(지라API_정보.getEndpoint().getIssue().getFull().getSubtask(), 이슈_키_또는_아이디);
                String endpoint = subTaskEndpoint + "&startAt=" + 검색_시작_지점 + "&maxResults=" + 최대_검색수;

                지라이슈조회_데이터 서브테스크_조회결과
                        = 지라유틸.getWithRawData(webClient, endpoint, 지라이슈조회_데이터.class).block();

                if (서브테스크_조회결과 == null) {
                    String error = String.format("클라우드 지라(%s) :: 이슈 키(%s) :: 서브테스크 조회 결과가 Null입니다.", 서버정보.getUri(), 이슈_키_또는_아이디);
                    log.info(error);

                    throw new Exception(error);
                }
                else if (서브테스크_조회결과.getIssues() == null || 서브테스크_조회결과.getIssues().size() == 0) {
                    log.info("클라우드 지라({}) :: 이슈 키({}) :: 서브테스크 이슈 목록이 없습니다.", 서버정보.getUri(), 이슈_키_또는_아이디);
                    return 서브테스크_목록;
                }

                서브테스크_조회결과.getIssues().forEach(하위이슈 -> 하위이슈.setUpperKey(이슈_키_또는_아이디));
                서브테스크_목록.addAll(서브테스크_조회결과.getIssues());

                검색_시작_지점 += 최대_검색수;
                if (검색_시작_지점 >= 서브테스크_조회결과.getTotal()) {
                    break;
                }
            }

            return 서브테스크_목록;
        }
        catch (Exception e) {
            String error = ErrorLogUtil.exceptionLoggingAndReturn(e, this.getClass().getName(),
                    "클라우드 지라("+ 서버정보.getUri() +") :: 이슈 키("+ 이슈_키_또는_아이디+ ") :: 서브테스크_가져오기에 실패하였습니다.");
            throw new Exception(error);
        }
    }


    @Override
    public List<지라이슈_데이터> 증분서브테스크_가져오기(서버정보_데이터 서버정보, String 이슈_키_또는_아이디,String startDate,String endDate) {

        WebClient webClient = 지라유틸.클라우드_통신기_생성(서버정보.getUri(), 서버정보.getUserId(), 서버정보.getPasswordOrToken());

        int 검색_시작_지점 = 0;
        int 최대_검색수 = 지라API_정보.getParameter().getMaxResults();
        boolean isLast = false;

        List<지라이슈_데이터> 서브테스크_목록 = new ArrayList<>(); // 이슈 저장

        try {
            while (!isLast) {

                String subtaskEndpoint = 지라API_정보.getEndpoint().getIssue().getIncrement().getSubtask();

                if(startDate!=null&&endDate!=null){
                    subtaskEndpoint = 지라API_정보.getEndpoint().getIssue().getIncrement().getSubtaskManualDate();
                    subtaskEndpoint = subtaskEndpoint.replace("{수동날짜}","updated > '"+startDate+"' AND updated < '"+getNextDate(endDate)+"'");
                }

                String endpoint = 지라API_정보.이슈키_대체하기(subtaskEndpoint, 이슈_키_또는_아이디)
                        + "&startAt=" + 검색_시작_지점 + "&maxResults=" + 최대_검색수;

                지라이슈조회_데이터 서브테스크_조회결과
                        = 지라유틸.getWithRawData(webClient, endpoint, 지라이슈조회_데이터.class).block();


                // 증분 데이터가 없는 경우
                if (서브테스크_조회결과 == null) {
                    return null;
                }
                else if (서브테스크_조회결과.getIssues() == null || 서브테스크_조회결과.getIssues().size() == 0) {
                    return null;
                }

                서브테스크_조회결과.getIssues().forEach(하위이슈 -> 하위이슈.setUpperKey(이슈_키_또는_아이디));
                서브테스크_목록.addAll(서브테스크_조회결과.getIssues());

                if (서브테스크_조회결과.getTotal() == 서브테스크_목록.size()) {
                    isLast = true;
                } else {
                    검색_시작_지점 += 최대_검색수;
                }
            }

            return 서브테스크_목록;
        }
        catch (Exception e) {
            ErrorLogUtil.exceptionLogging(e, this.getClass().getName(),
                    "클라우드 지라("+ 서버정보.getUri() +") :: 이슈 키("+ 이슈_키_또는_아이디+ ") :: 증분_서브테스크_가져오기에 실패하였습니다.");
            return null;
        }
    }

    public CloudJiraIssueFieldDTO validateAndAddFields(IssueCreationFieldsDTO IssueCreationFieldsDTO,
                                                       Map<String, CloudJiraIssueCreationFieldMetadata.FieldMetadata> fieldMetadataMap,
                                                       서버정보_데이터 서버정보, String projectKeyOrId, String issueTypeId) {

        if (fieldMetadataMap == null) {
            String 에러로그 = "필드검증_및_추가하기 필드_메타데이터_목록 Null 오류 클라우드 지라(" + 서버정보.getUri() + ") ::  프로젝트 :: "+projectKeyOrId+
                    " :: 이슈유형 :: " + issueTypeId+ " :: 생성 필드 :: "+ IssueCreationFieldsDTO.toString();
            log.error(에러로그);
            throw new IllegalArgumentException(에러로그);
        }

        CloudJiraIssueFieldDTO CloudJiraIssueFieldDTO = new CloudJiraIssueFieldDTO();

        if (fieldMetadataMap.containsKey(지라API_정보.getMetadata().getFields().getProject())
                && IssueCreationFieldsDTO.getProject() != null) {

            CloudJiraIssueFieldDTO.setProject(IssueCreationFieldsDTO.getProject());
            fieldMetadataMap.remove(지라API_정보.getMetadata().getFields().getProject());
        }

        if (fieldMetadataMap.containsKey(지라API_정보.getMetadata().getFields().getIssuetype())
                && IssueCreationFieldsDTO.getIssuetype() != null) {

            CloudJiraIssueFieldDTO.setIssuetype(IssueCreationFieldsDTO.getIssuetype());
            fieldMetadataMap.remove(지라API_정보.getMetadata().getFields().getIssuetype());
        }

        if (fieldMetadataMap.containsKey(지라API_정보.getMetadata().getFields().getSummary())
                && IssueCreationFieldsDTO.getSummary() != null) {

            CloudJiraIssueFieldDTO.setSummary(IssueCreationFieldsDTO.getSummary());
            fieldMetadataMap.remove(지라API_정보.getMetadata().getFields().getSummary());
        }

        if (fieldMetadataMap.containsKey(지라API_정보.getMetadata().getFields().getDescription())
                && IssueCreationFieldsDTO.getDescription() != null) {

            CloudJiraIssueFieldDTO.setDescription(내용_변환(IssueCreationFieldsDTO.getDescription()));
            fieldMetadataMap.remove(지라API_정보.getMetadata().getFields().getDescription());
        }

        if (fieldMetadataMap.containsKey(지라API_정보.getMetadata().getFields().getReporter())) {
            /***
             * reporter 필드는 필수 여부와 상관없이 추가하지 않아도 API를 요청한 사용자로 자동으로 추가된다.
             ***/
            fieldMetadataMap.remove(지라API_정보.getMetadata().getFields().getReporter());
        }

        if (fieldMetadataMap.containsKey(지라API_정보.getMetadata().getFields().getPriority())
                && IssueCreationFieldsDTO.getPriority() != null) {

            CloudJiraIssueFieldDTO.setPriority(IssueCreationFieldsDTO.getPriority());
            fieldMetadataMap.remove(지라API_정보.getMetadata().getFields().getPriority());
        }

        if (fieldMetadataMap.containsKey(지라API_정보.getMetadata().getFields().getDuedate())
                && IssueCreationFieldsDTO.getDueDate() != null) {

            String 종료기한일자 = DateUtils.formatDate(IssueCreationFieldsDTO.getDueDate(), "yyyy-MM-dd");
            CloudJiraIssueFieldDTO.setDuedate(종료기한일자);
            fieldMetadataMap.remove(지라API_정보.getMetadata().getFields().getDuedate());
        }

        /***
         * A-RMS에서 지원하는 필드 메타데이터 목록을 모두 remove 시킨 후 필수로 지정된 필드가 있을 경우 오류 반환
         ***/
        fieldMetadataMap.forEach((key, 필드_메타) -> {
            if (필드_메타.isRequired()) {
                String 에러로그 = "필수 필드 확인 및 추가 중 오류 [" + key + "] 필드가 필수로 지정되어있습니다. " +
                        "A-RMS에서 지원하지 않는 필드입니다. ";
                String 관련정보 = "클라우드 지라(" + 서버정보.getUri() + ") ::  프로젝트 :: "+projectKeyOrId+
                        " :: 이슈유형 :: " + issueTypeId+ " :: 생성 필드 :: "+ IssueCreationFieldsDTO.toString();
                log.error(에러로그 + 관련정보);
                throw new IllegalArgumentException(에러로그 + 관련정보);
            }
        });

        return CloudJiraIssueFieldDTO;
    }

}
